package net.es.oscars.policy.config;

import java.io.IOException;
import java.io.StreamTokenizer;
import java.util.HashMap;
import java.util.List;

import org.apache.log4j.Logger;

import net.es.oscars.policy.PolicyCore;
import net.es.oscars.policy.database.LTypeDAO;
import net.es.oscars.policy.database.OperatorDAO;
import net.es.oscars.policy.database.Rule;
import net.es.oscars.policy.database.RuleDAO;
import net.es.oscars.policy.database.RuleVar;

public class ParserUtil {
    private Logger log;
    private ConfigUtil configUtil;
    private HashMap<String, RuleVar> varNameMap;
    private String dbname;
    private RuleDAO ruleDAO;
    private LTypeDAO ltypeDAO;
    private OperatorDAO operatorDAO;
    private RuleSetParser ruleSetParser;
    
    static final public int CLASS_MODE = 1;
    static final public int RULESET_MODE = 2;
    static final public int MEMBERS_MODE = 3;
    static final public int MATCHANY_MODE = 4;
    static final public int MATCHALL_MODE = 5;
    static final public int VAR_MODE = 6;
    static final public int VAR_INPUT_MODE = 7;
    static final public int VAR_OUTPUT_MODE = 8;
    //must keep these after MATCH*_MODE
    static final public int VAR_INPUT_MATCHANY_MODE = 9;
    static final public int VAR_INPUT_MATCHALL_MODE = 10;
    
    static final public int POSITION_TOKEN = 1;
    static final public int LIST_TOKEN = 2;
    static final public int ENUM_TOKEN = 3;
    
    public ParserUtil(){
        this.log = Logger.getLogger(this.getClass());
        this.configUtil = new ConfigUtil();
        this.dbname = PolicyCore.getInstance().getPolicyDbName();
        this.ruleDAO = new RuleDAO(this.dbname);
        this.ltypeDAO = new LTypeDAO(this.dbname);
        this.operatorDAO = new OperatorDAO(this.dbname);
        this.varNameMap = new HashMap<String, RuleVar>();
        this.ruleSetParser = new RuleSetParser();
        this.ruleSetParser.init(this.configUtil, this.ruleDAO, 
                this.ltypeDAO, this.operatorDAO);
        this.ruleSetParser.setVarNameMap(this.varNameMap);
    }
    
    public static String readToken(StreamTokenizer tokenizer) {
        //read next token
        String token = null;
        if(tokenizer.ttype == StreamTokenizer.TT_WORD){
            token = tokenizer.sval;
        }else if(tokenizer.ttype == '"'|| tokenizer.ttype == '\''){
            token = tokenizer.sval;
        }else if(tokenizer.ttype == StreamTokenizer.TT_NUMBER){
            token = tokenizer.nval + "";
        }else{
            token = "" + (char)tokenizer.ttype;
        }
        return token;
    }
    
    static public Rule combineRule(Rule rule1, Rule rule2, String ruleOp, 
            LTypeDAO ltypeDAO, OperatorDAO operatorDAO, RuleDAO ruleDAO){
        Rule combinedRule = new Rule();
        combinedRule.setLType(ltypeDAO.queryByParam("name", "RULE"));
        combinedRule.setLValue(rule1.getId()+"");
        combinedRule.setOperator(operatorDAO.findByTypeName("RULE", ruleOp));
        combinedRule.setRValue(rule2.getId() + "");
        if(combinedRule.getLValue() != null && combinedRule.getRValue() != null){
            Rule existingCombRule = ruleDAO.findMatchingRule(combinedRule);
            if(existingCombRule != null){
                combinedRule = existingCombRule;
            }else{
                ruleDAO.create(combinedRule);
            }
        }
        
        return combinedRule;
    }
    
    public boolean parseMatchBlock(StreamTokenizer tokenizer, String token, String ruleOp, List<Integer> modeStack, List<Rule> ruleStack, RuleParser ruleParser) throws PolicyParserException, IOException{
        boolean openBracket = false;
        String normalizedToken = token;
        int tokenType = 0;
        if(token.matches(".+\\-\\d+$")){
            tokenType = POSITION_TOKEN;
            normalizedToken = normalizedToken.replaceAll("\\-\\d+$", "-#N#");
        }else if(token.endsWith("-all")){
            tokenType = LIST_TOKEN;
        }else if(token.endsWith("-any")){
            tokenType = ENUM_TOKEN;
        }
        
        if(token.equals("match-any")){
            openBracket = true;
            modeStack.add(0, (modeStack.get(0) > MATCHALL_MODE ? VAR_INPUT_MATCHANY_MODE : MATCHANY_MODE));
            ruleStack.add(0, null);
        }else if(token.equals("match-all")){
            openBracket = true;
            modeStack.add(0, (modeStack.get(0) > MATCHALL_MODE ? VAR_INPUT_MATCHALL_MODE : MATCHALL_MODE));
            ruleStack.add(0, null);
        }else if("}".equals(token)){
            modeStack.remove(0);
            
            //nested match
            String combinedOp = "OR";
            if(MATCHALL_MODE == modeStack.get(0) || VAR_INPUT_MATCHALL_MODE == modeStack.get(0)){
                combinedOp = "AND";
            }
            Rule rule = ruleStack.remove(0);
            if(ruleStack.get(0) == null){
                ruleStack.remove(0);
                ruleStack.add(0, rule);
            }else{
                Rule combinedRule = ParserUtil.combineRule(ruleStack.get(0), 
                        rule, combinedOp, this.ltypeDAO, this.operatorDAO, 
                        this.ruleDAO);
                ruleStack.remove(0);
                ruleStack.add(0, combinedRule);
            }
        }else{
            ruleParser.parseRule(tokenizer, ruleStack, token, 
                    normalizedToken, tokenType, ruleOp);
        }
        
        return openBracket;
    }
   
    public VarParser initVarBlock(StreamTokenizer tokenizer) throws PolicyParserException, IOException {
        String varType = ParserUtil.readToken(tokenizer);
        if(!this.configUtil.getVarParsers().containsKey(varType)){
            throw new PolicyParserException("Unknown var type '" + 
                    varType + "' at " + tokenizer.lineno());
        }
        VarParser currVarParser = this.configUtil.getVarParser(varType);
        currVarParser.init(this.configUtil, this.ruleDAO, this.ltypeDAO, this.operatorDAO);
        tokenizer.nextToken();
        if(!":".equals(ParserUtil.readToken(tokenizer))){
            throw new PolicyParserException("Expected ':' but found '" + 
                    ParserUtil.readToken(tokenizer) + "' at " + tokenizer.lineno());
        }
        tokenizer.nextToken();
        String varName = varType+":"+ParserUtil.readToken(tokenizer);
        if(this.varNameMap.containsKey(varName)){
            throw new PolicyParserException("Variable '" + varName + 
                    "' defined twice'" + "' at " + tokenizer.lineno());
        }
        currVarParser.setName(varName);
        
        return currVarParser;
    }

    /**
     * @return the ruleSetParser
     */
    public RuleSetParser getRuleSetParser() {
        return this.ruleSetParser;
    }

    /**
     * @return the varNameMap
     */
    public HashMap<String, RuleVar> getVarNameMap() {
        return this.varNameMap;
    }

    /**
     * @return the configUtil
     */
    public ConfigUtil getConfigUtil() {
        return this.configUtil;
    }
}
