package net.es.oscars.policy.config;

import jargs.gnu.CmdLineParser;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import org.apache.log4j.Logger;
import org.hibernate.Session;

import net.es.oscars.policy.PolicyCore;
import net.es.oscars.policy.database.*;
import net.es.oscars.policy.rmi.PolicyRmiClient;


/**
 * Parses a configuration specified by the command line and loads it
 * into a MySQL database.
 * 
 * @author Andrew Lake <alake@internet2.edu>
 *
 */
public class PolicyFileParser {
    private Logger log;
    private File policyFile;
    private String dbname;
    private PolicyCore core;
    private PolicyClassDAO policyClassDAO;
    private RuleSetDAO ruleSetDAO;
    private RuleDAO ruleDAO;
    private LTypeDAO ltypeDAO;
    private OperatorDAO operatorDAO;
    private PolicyClassCriterionDAO criteriaDAO;
    private ConfigUtil configUtil;
    private boolean verbose;

    final private int CLASS_MODE = 1;
    final private int RULESET_MODE = 2;
    final private int MEMBERS_MODE = 3;
    final private int MATCHANY_MODE = 4;
    final private int MATCHALL_MODE = 5;
    final private int POSITION_TOKEN = 1;
    final private int LIST_TOKEN = 2;
    final private int ENUM_TOKEN = 3;
    final private String[] tokenTypes = {"","XPATH_LIST", "XPATH_ENUM"};
    
    final private List<String> stringOps = Arrays.asList(new String[]{
            "==", "!=","startsWith", "endsWith", "contains"});
    final private List<String> numberOps = Arrays.asList(new String[]{
            "==", "!=","<",">","<=",">="});
    final private List<String> booleanOps = Arrays.asList(new String[]{
            "==", "!="});
    final private List<String> dateOps = Arrays.asList(new String[]{
            "==", "!=","<",">","<=",">="});
    
    public PolicyFileParser(String filename, Boolean verbose){
        this.log = Logger.getLogger(this.getClass());
        this.core = PolicyCore.getInstance();
        this.policyFile = new File(filename);
        this.dbname = this.core.getPolicyDbName();
        this.configUtil = new ConfigUtil();
        this.verbose = (verbose != null ? verbose : false);
    }
    
    public void parse() throws Exception{
        Session sess = this.core.getPolicySession();
        sess.beginTransaction();
        try{
            this.policyClassDAO = new PolicyClassDAO(this.dbname);
            this.ruleSetDAO = new RuleSetDAO(this.dbname);
            this.ruleDAO = new RuleDAO(this.dbname);
            this.ltypeDAO = new LTypeDAO(this.dbname);
            this.operatorDAO = new OperatorDAO(this.dbname);
            this.criteriaDAO = new PolicyClassCriterionDAO(this.dbname);
            this.parseFile();
            sess.getTransaction().commit();
        }catch(Exception e){
            sess.getTransaction().rollback();
            throw e;
        }
    }
    
    private void parseFile() throws IOException, PolicyParserException{
        if(this.verbose){
            System.out.println("Initializing parser...");
        }
        FileReader reader= new FileReader(policyFile);
        HashMap<String, PolicyClass> policyClassMap = new HashMap<String, PolicyClass>();
        HashMap<String, RuleSet> ruleSetMap = new HashMap<String, RuleSet>();
        HashMap<String, String> policyParentMap = new HashMap<String, String>();
        HashMap<String, String> policyRuleSetMap = new HashMap<String, String>();
        HashMap<String,List<PolicyClassCriterion>> policyCriteriaMap = new HashMap<String,List<PolicyClassCriterion>>();
        ArrayList<Integer> modeStack = new ArrayList<Integer>(); 
        StreamTokenizer tokenizer = new StreamTokenizer(reader);
        boolean openBracket = false;
        PolicyClass currPolicyClass = null;
        RuleSet currRuleSet = null;
        ArrayList<Rule> ruleStack = new ArrayList<Rule>();
        
        //initialize RulSetMap with default rule set
        ruleSetMap.put("default", ruleSetDAO.queryByParam("name", "default"));
        
        //initialize syntax table
        tokenizer.commentChar('#');
        tokenizer.eolIsSignificant(false);
        tokenizer.lowerCaseMode(false);
        //includes <,=,>
        tokenizer.wordChars('<','>');
        tokenizer.wordChars('!','!');
        if(this.verbose){
            System.out.println("Parser initialized.");
        }
        
        if(this.verbose){
            System.out.println("Parsing " + this.policyFile);
        }
        while(tokenizer.nextToken() != StreamTokenizer.TT_EOF){
            String token = this.readToken(tokenizer);
            /* look for open bracket */
            if(openBracket && "{".equals(token)){
                openBracket = false;
                continue;
            }else if(openBracket){
                throw new PolicyParserException("Missing { near line " + tokenizer.lineno());
            }
            
            /* determine current mode */
            if(modeStack.isEmpty()){
                //init mode
                openBracket = true;
                if(token.equals("policy-class")){
                    tokenizer.nextToken();
                    String className = this.readToken(tokenizer);
                    PolicyClass policyClass = policyClassDAO.queryByParam("name", className);
                    if(policyClass == null){
                        policyClass = new PolicyClass();
                        policyClass.setName(className);
                    }else{
                        policyClass.setParent(null); //clear out old value
                    }
                    if(policyClassMap.containsKey(className)){
                        throw new PolicyParserException("Policy class " + className + " defined twice at line " + tokenizer.lineno());
                    }
                    policyClassMap.put(className, policyClass);
                    currPolicyClass = policyClass;
                    modeStack.add(0, CLASS_MODE);
                }else if(token.equals("rule-set")){
                    tokenizer.nextToken();
                    String setName = this.readToken(tokenizer);
                    if("default".equals(setName)){
                        throw new PolicyParserException("Rule-set 'default' is " +
                                "a reserved name. Please rename the rule-set " +
                                "at line " + tokenizer.lineno());
                    }
                    RuleSet ruleSet = ruleSetDAO.queryByParam("name", setName);
                    if(ruleSet == null){
                        ruleSet = new RuleSet();
                        ruleSet.setName(setName);
                    }
                    if(ruleSetMap.containsKey(setName)){
                        throw new PolicyParserException("Rule set " + setName + " defined twice at line " + tokenizer.lineno());
                    }
                    ruleSetMap.put(setName, ruleSet);
                    currRuleSet = ruleSet;
                    modeStack.add(0, RULESET_MODE);
                    ruleStack.add(0, null);
                }else{
                    throw new PolicyParserException("Invalid token '" + token + "' at line " + tokenizer.lineno());
                }
            }else if(modeStack.get(0) == CLASS_MODE){
                if(token.equals("members")){
                    openBracket = true;
                    modeStack.add(0, MEMBERS_MODE);
                }else if(token.equals("parent")){
                    tokenizer.nextToken();
                    String parentName = this.readToken(tokenizer);
                    if(policyParentMap.containsKey(currPolicyClass.getName())){
                        throw new PolicyParserException("Cannot define multiple parents at line " + tokenizer.lineno());
                    }
                    policyParentMap.put(currPolicyClass.getName(), parentName);
                    tokenizer.nextToken();
                    if(!";".equals(this.readToken(tokenizer))){
                        throw new PolicyParserException("Missing ';' at line " + tokenizer.lineno());
                    }
                }else if(token.equals("rule-set")){
                    tokenizer.nextToken();
                    String setName = this.readToken(tokenizer);
                    if(policyRuleSetMap.containsKey(currPolicyClass.getName())){
                        throw new PolicyParserException("Cannot define multiple rule-sets at line " + tokenizer.lineno());
                    }
                    policyRuleSetMap.put(currPolicyClass.getName(), setName);
                    tokenizer.nextToken();
                    String allowOrDeny = this.readToken(tokenizer);
                    if("allow".equals(allowOrDeny)){
                        currPolicyClass.setAllowOnMatch(true);
                    }else if("deny".equals(allowOrDeny)){
                        currPolicyClass.setAllowOnMatch(false);
                    }else{
                        throw new PolicyParserException("Expected 'allow' or 'deny' but found " + allowOrDeny + " at line " + tokenizer.lineno());
                    }
                    tokenizer.nextToken();
                    if(!";".equals(this.readToken(tokenizer))){
                        throw new PolicyParserException("Missing ';' at line " + tokenizer.lineno());
                    }
                }else if("}".equals(token)){
                    currPolicyClass = null;
                    modeStack.remove(0);
                }else{
                    throw new PolicyParserException("Invalid token '" + token + "' in 'policy-class' block at line " + tokenizer.lineno());
                }
            }else if(modeStack.get(0) == RULESET_MODE){
                if(token.equals("match-any")){
                    openBracket = true;
                    modeStack.add(0, MATCHANY_MODE);
                    ruleStack.add(0, null);
                }else if(token.equals("match-all")){
                    openBracket = true;
                    modeStack.add(0, MATCHALL_MODE);
                    ruleStack.add(0, null);
                }else if("}".equals(token) && ruleStack.size() == 1){
                    currRuleSet.setRule(ruleStack.get(0));
                    ruleStack.remove(0);
                    ruleSetDAO.create(currRuleSet);
                    if(this.verbose){
                        System.out.println("Created rule set " + currRuleSet.getName());
                    }
                    currRuleSet = null;
                    modeStack.remove(0);
                }else{
                    throw new PolicyParserException("Invalid token '" + token + "' in 'rule-set' block at line " + tokenizer.lineno());
                }
            }else if(modeStack.get(0) == MEMBERS_MODE){
                if(token.equals("name")){
                    tokenizer.nextToken();
                    this.createCriteria(tokenizer, "NAME", currPolicyClass, policyCriteriaMap);
                    tokenizer.nextToken();
                    if(!";".equals(this.readToken(tokenizer))){
                        throw new PolicyParserException("Missing ';' at line " + tokenizer.lineno());
                    }
               }else if(token.equals("role")){
                   tokenizer.nextToken();
                   this.createCriteria(tokenizer, "ATTR_ROLE", currPolicyClass, policyCriteriaMap);
                   tokenizer.nextToken();
                   if(!";".equals(this.readToken(tokenizer))){
                       throw new PolicyParserException("Missing ';' at line " + tokenizer.lineno());
                   }
               }else if("}".equals(token)){
                    modeStack.remove(0);
               }else{
                   throw new PolicyParserException("Invalid token '" + token + "' in 'members' block at line " + tokenizer.lineno());
               }
            }else if(modeStack.get(0) == MATCHANY_MODE){
                openBracket = this.parseMatchBlock(tokenizer, token, "OR", modeStack, ruleStack);
            }else if(modeStack.get(0) == MATCHALL_MODE){
                openBracket = this.parseMatchBlock(tokenizer, token, "AND", modeStack, ruleStack);
            }
        }
        if(!modeStack.isEmpty()){
            throw new PolicyParserException("Unexpected EOF. Make sure you are not missing any closing brackets.");
        }
        
        //Initialize list of policyClasses to delete with all existing classes
        List<PolicyClass> existingPolicyClasses = policyClassDAO.list();
        HashMap<String, PolicyClass> policyDeleteMap = new HashMap<String, PolicyClass>();
        for(PolicyClass policyClass : existingPolicyClasses){
            policyDeleteMap.put(policyClass.getName(), policyClass);
        }
        
        //Initialize list of policyClassCriteria to delete with all existing classes
        List<PolicyClassCriterion> existingCriteria = criteriaDAO.list();
        HashMap<String,PolicyClassCriterion> criteriaDeleteMap = new HashMap<String,PolicyClassCriterion>();
        for(PolicyClassCriterion crit : existingCriteria){
            criteriaDeleteMap.put(critKey(crit), crit);
        }
        
        //Save new policyClasses now that all rule sets and parents have been created
        for(String className : policyClassMap.keySet()){
            if(!ruleSetMap.containsKey(policyRuleSetMap.get(className))){
                throw new PolicyParserException("Unable to find rule-set named '"+ 
                        policyRuleSetMap.get(className) + "' for policy class '" + className + "'");
            }
            policyClassMap.get(className).setRuleSet(ruleSetMap.get(policyRuleSetMap.get(className)));
            
            if(policyParentMap.containsKey(className)){
                if(!policyClassMap.containsKey(policyParentMap.get(className))){
                    throw new PolicyParserException("Unable to find parent " +
                            "policy class named '"+ policyParentMap.get(className) + 
                            "' for policy class '" + className + "'");
                }
                policyClassMap.get(className).setParent(policyClassMap.get(policyParentMap.get(className)));
            }
            this.policyClassDAO.create(policyClassMap.get(className));
            if(this.verbose){
                System.out.println("Created policy class " + className);
            }
            
            //remove this class from list
            if(policyDeleteMap.containsKey(className)){
                policyDeleteMap.remove(className);
            }
            
            //save criteria
            if("default".equals(className)){
                continue;
            }else if(!policyCriteriaMap.containsKey(className)){
                throw new PolicyParserException("Policy class '"+className+"' does not contain any members");
            }

            for(PolicyClassCriterion crit : policyCriteriaMap.get(className)){
                if(criteriaDeleteMap.containsKey(critKey(crit))){
                   criteriaDeleteMap.remove(critKey(crit));
                }
                criteriaDAO.create(crit);
            }
        }
        
        //delete all unused criteria
        for(String critKey : criteriaDeleteMap.keySet()){
           criteriaDAO.remove(criteriaDeleteMap.get(critKey));
        }
        
        //delete all unused policy classes
        for(String className : policyDeleteMap.keySet()){
            policyClassDAO.remove(policyDeleteMap.get(className));
            if(this.verbose){
                System.out.println("Deleted policy class" + className);
            }
        }
        
        
    }
    
    private boolean parseMatchBlock(StreamTokenizer tokenizer, String token, String ruleOp, List<Integer> modeStack, List<Rule> ruleStack) throws PolicyParserException, IOException{
        boolean openBracket = false;
        String normalizedToken = token;
        int tokenType = 0;
        if(token.matches(".+\\-\\d+$")){
            tokenType = POSITION_TOKEN;
            normalizedToken = normalizedToken.replaceAll("\\-\\d+$", "-#N#");
        }else if(token.endsWith("-all")){
            tokenType = LIST_TOKEN;
        }else if(token.endsWith("-any")){
            tokenType = ENUM_TOKEN;
        }
        
        if(token.equals("match-any")){
            openBracket = true;
            modeStack.add(0, MATCHANY_MODE);
            ruleStack.add(0, null);
        }else if(token.equals("match-all")){
            openBracket = true;
            modeStack.add(0, MATCHALL_MODE);
            ruleStack.add(0, null);
        }else if(this.configUtil.getStringFields().contains(normalizedToken)){
            String lType = (tokenType > POSITION_TOKEN) ? this.tokenTypes[tokenType-1] : "XPATH_NODE";
            this.parseRule(tokenizer, ruleStack, lType, 
                    this.prepareXPath(token, normalizedToken, tokenType), "STRING", 
                    this.stringOps, ruleOp);
        }else if(this.configUtil.getNumberFields().contains(normalizedToken)){
            String lType = (tokenType > POSITION_TOKEN) ? this.tokenTypes[tokenType-1] : "XPATH_NUMBER";
            this.parseRule(tokenizer, ruleStack, lType, 
                    this.prepareXPath(token, normalizedToken, tokenType), "NUMBER", 
                    this.numberOps, ruleOp);
        }else if(this.configUtil.getBooleanFields().contains(normalizedToken)){
            String lType = (tokenType > POSITION_TOKEN) ? this.tokenTypes[tokenType-1] : "XPATH_BOOLEAN";
            this.parseRule(tokenizer, ruleStack, lType, 
                    this.prepareXPath(token, normalizedToken, tokenType), "BOOLEAN", 
                    this.booleanOps, ruleOp);
        }else if(this.configUtil.getDateFields().contains(normalizedToken)){
            String lType = (tokenType > POSITION_TOKEN) ? this.tokenTypes[tokenType-1] : "XPATH_NUMBER";
            this.parseRule(tokenizer, ruleStack, lType, 
                    this.prepareXPath(token, normalizedToken, tokenType), "DATE", 
                    this.dateOps, ruleOp);
        }else if("action".equals(token)){
            this.parseRule(tokenizer, ruleStack, "ACTION", "*", "STRING", this.stringOps, ruleOp);
        }else if("}".equals(token)){
            modeStack.remove(0);
            
            //nested match
            String combinedOp = "OR";
            if(MATCHALL_MODE == modeStack.get(0)){
                combinedOp = "AND";
            }
            Rule rule = ruleStack.remove(0);
            if(ruleStack.get(0) == null){
                ruleStack.remove(0);
                ruleStack.add(0, rule);
            }else{
                Rule combinedRule = this.combineRule(ruleStack.get(0), rule, combinedOp);
                ruleStack.remove(0);
                ruleStack.add(0, combinedRule);
            }
        }else{
            throw new PolicyParserException("Invalid token '" + token + "' in 'match-all' block at line " + tokenizer.lineno());
        }
        
        return openBracket;
    }
   

    private void parseRule(StreamTokenizer tokenizer, List<Rule> ruleStack, 
            String lType, String lValue, String rType, List<String> ops, 
            String ruleOp) throws PolicyParserException, IOException{
        tokenizer.nextToken();
        this.log.info("parseRule." + lType + ".start");
        String op = this.readToken(tokenizer);
        if(op == null || (!ops.contains(op))){
            throw new PolicyParserException("Invalid operator '" + op + "' at line " + tokenizer.lineno());
        }
        tokenizer.nextToken();
        String value = this.readToken(tokenizer);
        //adjust value
        if("ACTION".equals(lType) && this.configUtil.getActionMap().containsKey(value)){
            value = this.configUtil.getActionMap().get(value);
        }
        //validate date format
        if("DATE".equals(rType) && (!this.configUtil.validateDateField(value))){
            throw new PolicyParserException("Invalid date '" + value + "' at line " + tokenizer.lineno());
        }
        //check if special variable
        if(lValue.matches("#.+#")){
            lType = "SPECIAL_VAR";
            lValue = lValue.replaceAll("#","");
        }
        tokenizer.nextToken();
        if(!";".equals(this.readToken(tokenizer))){
            throw new PolicyParserException("Missing ';' at line " + tokenizer.lineno());
        }
        
        Rule rule = new Rule();
        rule.setLType(ltypeDAO.queryByParam("name", lType));
        rule.setLValue(lValue);
        rule.setOperator(operatorDAO.findByTypeName(rType, op));
        rule.setRValue(value);
        Rule existingRule = ruleDAO.findMatchingRule(rule);
        if(existingRule != null){
            rule = existingRule;
        }else{
            ruleDAO.create(rule);
        }
        
        //look at last rule
        if(ruleStack.get(0) == null){
            ruleStack.remove(0);
            ruleStack.add(0, rule);
        }else {
            Rule combinedRule = this.combineRule(ruleStack.get(0), rule, ruleOp);
            ruleStack.remove(0);
            ruleStack.add(0, combinedRule);
        }
        this.log.info("parseRule." + lType + ".end");
    }
    
    private Rule combineRule(Rule rule1, Rule rule2, String ruleOp){
        Rule combinedRule = new Rule();
        combinedRule.setLType(ltypeDAO.queryByParam("name", "RULE"));
        combinedRule.setLValue(rule1.getId()+"");
        combinedRule.setOperator(operatorDAO.findByTypeName("RULE", ruleOp));
        combinedRule.setRValue(rule2.getId() + "");
        if(combinedRule.getLValue() != null && combinedRule.getRValue() != null){
            Rule existingCombRule = ruleDAO.findMatchingRule(combinedRule);
            if(existingCombRule != null){
                combinedRule = existingCombRule;
            }else{
                ruleDAO.create(combinedRule);
            }
        }
        
        return combinedRule;
    }
    private void createCriteria(StreamTokenizer tokenizer, String typeName, 
            PolicyClass policyClass, HashMap<String,List<PolicyClassCriterion>> policyCriteriaMap) 
                throws PolicyParserException {
        PolicyClassCriterionTypeDAO criteriaTypeDAO = new PolicyClassCriterionTypeDAO(this.dbname);
        String memberVal = this.readToken(tokenizer);
        PolicyClassCriterion criteria = criteriaDAO.findByClassTypeValue(policyClass.getId(), typeName, memberVal);
        if(criteria == null){
            criteria = new PolicyClassCriterion();
            criteria.setPolicyClass(policyClass);
            criteria.setType(criteriaTypeDAO.queryByParam("name", typeName));
            criteria.setValue(memberVal);
        }
        if(!policyCriteriaMap.containsKey(policyClass.getName())){
            policyCriteriaMap.put(policyClass.getName(), new ArrayList<PolicyClassCriterion>());
        }else if(policyCriteriaMap.get(policyClass.getName()).contains(criteria)){
            throw new PolicyParserException("Same member defined twice at line " + tokenizer.lineno());
        }
        policyCriteriaMap.get(policyClass.getName()).add(criteria);
    }

    private String readToken(StreamTokenizer tokenizer) {
      //read next token
        String token = null;
        if(tokenizer.ttype == StreamTokenizer.TT_WORD){
            token = tokenizer.sval;
        }else if(tokenizer.ttype == '"'|| tokenizer.ttype == '\''){
            token = tokenizer.sval;
        }else if(tokenizer.ttype == StreamTokenizer.TT_NUMBER){
            token = tokenizer.nval + "";
        }else{
            token = "" + (char)tokenizer.ttype;
        }
        return token;
    }
    
    public String critKey(PolicyClassCriterion crit){
        return crit.getPolicyClass().getName()+crit.getType().getName()+crit.getValue();
    }
    
    /**
     * @return the verbose
     */
    public boolean isVerbose() {
        return this.verbose;
    }

    /**
     * @param verbose the verbose to set
     */
    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    private String prepareXPath(String token, String normalizedToken, int tokenType) {
        String xPath = this.configUtil.getXPathMap().get(normalizedToken);
        if(tokenType == POSITION_TOKEN){
            String[] parts = token.split("-");
            xPath = xPath.replace("#N#", parts[parts.length-1]);
        }
        
        return xPath;
    }
    
    public static void printUsage(){
        System.out.println("Usage: policy-config [(-v|--verbose)(-h|--help)] <config-filename>");
    }
    
    public static void main(String[] args){
        CmdLineParser parser = new CmdLineParser();
        CmdLineParser.Option optVerbose = parser.addBooleanOption('v', "verbose");
        CmdLineParser.Option optHelp = parser.addBooleanOption('h', "help");
        
        try {
            parser.parse(args);
        } catch (Exception e) {
            System.err.println(e.getMessage());
            System.exit(2);
        }
        Boolean help = (Boolean)parser.getOptionValue(optHelp);
        if(help != null && help){
            printUsage();
            System.exit(0);
        }
        Boolean verbose = (Boolean)parser.getOptionValue(optVerbose);
        if(parser.getRemainingArgs() == null || parser.getRemainingArgs().length != 1){
            printUsage();
            System.exit(2);
        }
        PolicyRmiClient rmiClient = new PolicyRmiClient();
        try {
            rmiClient.init();
            File file = new File(parser.getRemainingArgs()[0]);
            rmiClient.loadPolicy(file.getAbsolutePath(), verbose);
        } catch (Exception e) {
            System.err.println(e.getMessage());
            e.printStackTrace();
            System.exit(2);
        }
        System.out.println("Policy configuration loaded.");
    }
}
